using System;
using System.Collections;
using System.Linq;
using System.Collections.Generic;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using UnityEngine;
using UnityEngine.UI;
using SimpleJSON;
using Object = UnityEngine.Object;


namespace CheesyFX
{
    public partial class LightMeUp : MVRScript
    {
        private bool _toggleState = true;
        public static List<HeadLight> headLights = new List<HeadLight>();
        public static Transform cameraTransform;
        
        private JSONStorableStringChooser lightChooser;
        public static JSONStorableBool editSymmetrical;
        public static JSONStorableBool autoFocus = new JSONStorableBool("Auto Focus Side Lights", false);

        private bool sideLightEnabled => headLights[0].enabled.val || headLights[2].enabled.val;
        
        private RectTransform versionRectTransform;
        private RectTransform buttonsTransform;
        private RectTransform toggleButtonTransform;
        
        private ColorBlock activeColorBlock;
        public static ColorBlock disabledColorBlock;

        private KeyCode toggleKey = KeyCode.KeypadPlus;
        private KeyCode toggleLeft = KeyCode.KeypadDivide;
        private KeyCode toggleCenter = KeyCode.KeypadMultiply;
        private KeyCode toggleRight = KeyCode.KeypadMinus;

        private List<object> UIElements = new List<object>();

        public static Color[] defaultColors = 
        {
            new Color(1f, 0.89f, 0.78f),
            new Color(1f,1f,1f),
            new Color(0.78f, 0.98f, 1f),
        };
        
        private Dictionary<Atom, bool> sceneLights = new Dictionary<Atom, bool>();
        private bool isSceneLit = true;

        public override void Init()
        {
            MacGruber.Utils.OnInitUI(CreateUIElement);;
            SelectCameraRig();
            disabledColorBlock = ColorBlock.defaultColorBlock;
            disabledColorBlock.normalColor *= .5f;
            disabledColorBlock.highlightedColor *= .5f;
            disabledColorBlock.pressedColor *= .5f;
            headLights = new List<HeadLight>()
            {
                new HeadLight(this, 0),
                new HeadLight(this, 1),
                new HeadLight(this, 2)
            };
            CreateMainHUD();
            
            lightChooser = new JSONStorableStringChooser(
                "LightChooser",
                headLights.Select(x => x.name).ToList(),
                "Center",
                "Light",
                x => CreateLightUI(lightChooser.choices.IndexOf(x))
                );
            CreatePopup(lightChooser);
            UIDynamic spacer = CreateSpacer(true);
            spacer.height = 10f;
            editSymmetrical = new JSONStorableBool("Edit Symmetrical (except color)", true);
            CreateToggle(editSymmetrical, true);
            CreateToggle(autoFocus, true);
            MacGruber.Utils.SetupTwinButton(this, "Swap L/R colors", SwapLeftRightColor, "ToggleSceneLights", ToggleSceneLighting, false);
            MacGruber.Utils.SetupInfoOneLine(this, "Light Settings", false);
            spacer = CreateSpacer(true);
            spacer.height = 45f;
            CreateLightUI(1);
            Toggle();
        }

        private void SelectCameraRig()
        {
            SuperController singleton = SuperController.singleton;
            if (singleton.isOpenVR)
            {
                cameraTransform = singleton.ViveCenterCamera.transform;
            }
            else if (singleton.isOVR)
            {
                cameraTransform = singleton.OVRCenterCamera.transform;
            }
            else if (singleton.IsMonitorOnly)
            {
                cameraTransform = singleton.MonitorCenterCamera.transform;
            }
        }

        void CreateLightUI(int id)
        {
            MacGruber.Utils.RemoveUIElements(this, UIElements);
            UIElements = headLights[id].CreateUI();
        }

        private void CreateMainHUD()
        {
            //.37f
            RectTransform panel = SuperController.singleton.GetAtomByUid("CoreControl").gameObject.transform.Find("WorldScaleAdjust/HUD/LowerHUDPivot/LowerHUDFlip/Scene Control Canvas/Panel").GetComponent<RectTransform>();
            versionRectTransform = panel.Find("VersionText").GetComponent<RectTransform>();
            versionRectTransform.anchorMax = new Vector2(.25f, 0f);
            headLights.ForEach(x => x. CreateHUDToggle(panel));
            
            toggleButtonTransform = manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
            toggleButtonTransform = Instantiate(toggleButtonTransform, panel);
            toggleButtonTransform.name = $"ToggleButton";
            toggleButtonTransform.anchorMax = new Vector2(.38f, 1);
            toggleButtonTransform.anchorMin = new Vector2(.25f, 0);
            toggleButtonTransform.offsetMax = new Vector2(0, -5f);
            toggleButtonTransform.offsetMin = new Vector2(0, 5f);
            Button button = toggleButtonTransform.GetComponent<Button>();
            Text buttonText = toggleButtonTransform.Find("Text").GetComponent<Text>();
            buttonText.text = "LightMeUp";
            button.onClick.AddListener(Toggle);
            toggleButtonTransform.gameObject.SetActive(false);
        }

        public void Update()
        {
            // cameraTransform.position.Print();
            if (Input.GetKeyDown(toggleKey))
            {
                if(Input.GetKey(KeyCode.RightControl) || Input.GetKey(KeyCode.LeftControl)) ToggleSceneLighting();
                else Toggle();
            }
            else if(Input.GetKeyDown(toggleLeft)) headLights[0].Toggle();
            else if(Input.GetKeyDown(toggleCenter)) headLights[1].Toggle();
            else if(Input.GetKeyDown(toggleRight)) headLights[2].Toggle();
            if (!autoFocus.val || !_toggleState || !sideLightEnabled) return;
            RaycastHit hit;
            Ray ray = new Ray(cameraTransform.position, cameraTransform.forward);
            if(Physics.Raycast(ray, out hit))
            {
                Vector3 focusPoint = hit.point;
                headLights[0].UpdateRotation(focusPoint);
                headLights[2].UpdateRotation(focusPoint);
            }
        }

        void Toggle()
        {
            toggleButtonTransform.gameObject.SetActive(_toggleState);
            _toggleState = !_toggleState;
            headLights.ForEach(x => x.EnabledCallback(_toggleState && x.enabled.val));
        }

        void SwapLeftRightColor()
        {
            HSVColor leftColor = headLights[0].color.val;
            headLights[0].color.val = headLights[2].color.val;
            headLights[2].color.val = leftColor;
        }
        
        
        void ToggleSceneLighting()
        {
            
            if(isSceneLit) sceneLights.Clear();
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                if (atom.type == "InvisibleLight")
                {
                    JSONStorableBool on = atom.GetStorableByID("Light").GetBoolJSONParam("on");
                    if (isSceneLit)
                    {
                        sceneLights.Add(atom, on.val);
                        on.val = false;
                    }
                    else on.val = sceneLights[atom];
                }
            }
            isSceneLit = !isSceneLit;
        }

        public void OnEnable()
        {
            headLights.ForEach(x => x.enabled.val = true); 
        }

        public void OnDisable()
        {
            headLights.ForEach(x => x.enabled.val = false);
            if(!isSceneLit) ToggleSceneLighting();
        }

        public void OnDestroy()
        {
            headLights.ForEach(x => x.Destroy());
            versionRectTransform.anchorMax = new Vector2(.37f, 0f);
            Destroy(toggleButtonTransform.gameObject);
            if(!isSceneLit) ToggleSceneLighting();
        }
    }

    public class HeadLight
    {
        private MVRScript script;

        private int id;
        public string name;
        
        public GameObject gameObject;
        private Light light;

        public JSONStorableBool enabled;
        
        private JSONStorableStringChooser type;
        private JSONStorableStringChooser renderMode;
        private JSONStorableStringChooser shadowResolution;
        private JSONStorableFloat intensity;
        
        private JSONStorableFloat spotAngle;
        private JSONStorableFloat range;
        private JSONStorableBool castShadows;
        private JSONStorableFloat shadowStrength;
        public JSONStorableColor color;
        
        
        private JSONStorableFloat offsetX;
        private JSONStorableFloat rotation;
        
        private Button button;
        private RectTransform buttonTransform;
        private string buttonText;
        private ColorBlock activeColor;

        public HeadLight(MVRScript script, int id)
        {
            this.script = script;
            this.id = id;
            
            float offsetMin, offsetMax, offsetDefault, rotationMin, rotationMax, rotationDefault;
            switch (id)
            {
                case 0:
                    offsetMin = -3f;
                    offsetMax = 0f;
                    offsetDefault = -.75f;
                    rotationDefault = 30f;
                    rotationMin = 0f;
                    rotationMax = 90f;
                    buttonText = "L";
                    name = "Left";
                    break;
                case 1:
                    offsetMin = -3f;
                    offsetMax = 3f;
                    offsetDefault = 0f;
                    rotationDefault = 0f;
                    rotationMin = -90f;
                    rotationMax = 90f;
                    buttonText = "C";
                    name = "Center";
                    break;
                default:
                    offsetMin = 0f;
                    offsetMax = 3f;
                    offsetDefault = .75f;
                    rotationDefault = -30f;
                    rotationMin = -90f;
                    rotationMax = 0f;
                    buttonText = "R";
                    name = "Right";
                    break;
            }
            Vector3 offset = new Vector3(offsetDefault, 0f, 0f);
            
            gameObject = new GameObject();
            Transform transform = gameObject.transform;
            transform.parent = LightMeUp.cameraTransform;
            transform.localPosition = new Vector3();
            transform.localRotation = Quaternion.identity;
            transform.localPosition = offset;
            
            light = gameObject.AddComponent<Light>();

            enabled = new JSONStorableBool("Enabled", true, EnabledCallback);
            type = new JSONStorableStringChooser("Type", Enum.GetNames(typeof(LightType)).ToList(), LightType.Spot.ToString(), "Type", TypeCallback);
            renderMode = new JSONStorableStringChooser("Render Mode", new List<string>{"Auto", "ForcePixel","ForceVertex"}, "Auto", "Render Mode", RenderModeCallback);
            shadowResolution = new JSONStorableStringChooser("Shadow Resolution", new List<string>{"FromQualitySettings, Low","Medium", "High", "VeryHigh"}, "FromQualitySettings", "Shadow Resolution");
            intensity = new JSONStorableFloat("Intensity", .5f, IntensityCallback, 0f, 8f, false);
            range = new JSONStorableFloat("Range", 5f, RangeCallback, 0f, 10f, false);
            shadowStrength = new JSONStorableFloat("Shadow Strength", 5f, ShadowStrengthCallback, 0f, 10f, false);
            spotAngle = new JSONStorableFloat("Spot Angle", 90f, SpotAngleCallback, 0f, 180f, false);
            castShadows = new JSONStorableBool("Cast Shadows", false, CastShadowsCallback);
            color = new JSONStorableColor("Color", LightMeUp.defaultColors[id].ToHSV(), ColorCallback);
            offsetX = new JSONStorableFloat("Horizontal Offset",offset.x, OffsetCallback, offsetMin, offsetMax);
            rotation = new JSONStorableFloat("Rotation",rotationDefault, RotationCallback, rotationMin, rotationMax);
            
            light.color = LightMeUp.defaultColors[id];
            range.val = 5f;
            light.intensity = intensity.defaultVal;
            light.spotAngle = spotAngle.defaultVal;
            light.type = LightType.Spot;
            gameObject.transform.localRotation = Quaternion.Euler(0f, rotationDefault, 0f);
            castShadows.val = id == 1;
        }

        public void Toggle()
        {
            enabled.val = !enabled.val;
        }

        public void EnabledCallback(bool b)
        {
            gameObject.SetActive(b);
            button.colors = b ? activeColor : LightMeUp.disabledColorBlock;
        }

        private void CastShadowsCallback(bool b)
        {
            light.shadows = b ? LightShadows.Hard : LightShadows.None;
        }

        void ShadowStrengthCallback(float f)
        {
            light.shadowStrength = f;
            if (LightMeUp.editSymmetrical.val && id != 1)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.shadowStrength = f;
                headLight.shadowStrength.val = f;
            }
        }

        void TypeCallback(string x)
        {
            light.type = (LightType)Enum.Parse(typeof(LightType), x);
            if (LightMeUp.editSymmetrical.val && id != 1)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.type = (LightType)Enum.Parse(typeof(LightType), x);
                headLight.type.valNoCallback = x;
            }
        }
        
        void RenderModeCallback(string x)
        {
            light.renderMode = (LightRenderMode)Enum.Parse(typeof(LightRenderMode), x);
            if (LightMeUp.editSymmetrical.val && id != 1)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.renderMode = (LightRenderMode)Enum.Parse(typeof(LightRenderMode), x);
                headLight.type.valNoCallback = x;
            }
        }

        void IntensityCallback(float f)
        {
            light.intensity = f;
            if (LightMeUp.editSymmetrical.val && id != 1)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.intensity = f;
                headLight.intensity.valNoCallback = f;
            }
        }
        
        void RangeCallback(float f)
        {
            light.range = f;
            if (LightMeUp.editSymmetrical.val && id != 1)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.range = f;
                headLight.range.valNoCallback = f;
            }
        }
        
        void SpotAngleCallback(float f)
        {
            light.spotAngle = f;
            if (LightMeUp.editSymmetrical.val && id != 1)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.spotAngle = f;
                headLight.spotAngle.valNoCallback = f;
            }
        }
        
        void OffsetCallback(float f)
        {
            gameObject.transform.localPosition = new Vector3(f, 0f, 0f);
            if (LightMeUp.editSymmetrical.val && id != 1)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.gameObject.transform.localPosition = new Vector3(-f, 0f, 0f);
                headLight.offsetX.valNoCallback = -f;
            }
        }

        void RotationCallback(float r)
        {
            Quaternion rotation = Quaternion.Euler(0f, r, 0f);
            gameObject.transform.localRotation = rotation;
            if (LightMeUp.editSymmetrical.val && id != 1)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                rotation = Quaternion.Euler(0f, -r, 0f);
                headLight.gameObject.transform.localRotation = rotation;
                headLight.rotation.valNoCallback = -r;
            } 
        }

        void ColorCallback(float h, float s, float v)
        {
            Color color = Color.HSVToRGB(h, s, v);
            light.color = color;
            activeColor.normalColor = color;
            activeColor.pressedColor = color;
            activeColor.highlightedColor = color;
            if (enabled.val) button.colors = activeColor;
        }

        public void UpdateRotation(Vector3 focusPoint)
        {
            Quaternion goal = new Quaternion();
            goal.SetLookRotation(focusPoint - gameObject.transform.position);
            Quaternion current = gameObject.transform.rotation;
            gameObject.transform.rotation = Quaternion.Lerp(current, goal, 4*Time.deltaTime);
            rotation.valNoCallback = gameObject.transform.localRotation.eulerAngles[1];
        }

        public void Destroy()
        {
            Object.Destroy(gameObject);
            Object.Destroy(buttonTransform.gameObject);
        }

        public List<object> CreateUI()
        {
            List<object> UIElements = new List<object>
            {
                script.CreatePopup(type),
                script.CreatePopup(renderMode),
                script.CreateSlider(intensity, false),
                script.CreateSlider(range),
                script.CreateSlider(spotAngle),
                
                script.CreateToggle(castShadows),
                // script.CreateSlider(shadowStrength),
                script.CreateColorPicker(color, true),
                
                
                script.CreateSlider(rotation, true),
                script.CreateSlider(offsetX, true)
            };
            return UIElements;
        }

        public void CreateHUDToggle(RectTransform parent)
        {
            buttonTransform = script.manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
            buttonTransform = Object.Instantiate(buttonTransform, parent);
            buttonTransform.name = $"HeadLight{id}";
            buttonTransform.anchorMax = new Vector2(.25f+.13f*(id+1)/3, 1);
            buttonTransform.anchorMin = new Vector2(.25f+.13f*id/3, 0);
            buttonTransform.offsetMax = new Vector2(0, -5f);
            buttonTransform.offsetMin = new Vector2(0, 5f);
            button = buttonTransform.GetComponent<Button>();
            buttonTransform.Find("Text").GetComponent<Text>().text = this.buttonText;
            
            // button.onClick.AddListener(() => headLight.CallAction("ToggleOn"));
            button.onClick.AddListener(Toggle);
            activeColor = ColorBlock.defaultColorBlock;
            var color1 = light.color;
            activeColor.normalColor = color1;
            activeColor.pressedColor = color1;
            activeColor.highlightedColor = color1;
            button.colors = activeColor;
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            enabled.Store(jc);
            color.Store(jc);
            type.Store(jc);
            renderMode.Store(jc);
            shadowResolution.Store(jc);
            intensity.Store(jc);
            spotAngle.Store(jc);
            range.Store(jc);
            castShadows.Store(jc);
            shadowStrength.Store(jc);
            offsetX.Store(jc);
            rotation.Store(jc);
            return jc;
        }

        public void LoadJSON(JSONClass jc)
        {
            enabled.Load(jc);
            color.Load(jc);
            type.Load(jc);
            renderMode.Load(jc);
            shadowResolution.Load(jc);
            intensity.Load(jc);
            spotAngle.Load(jc);
            range.Load(jc);
            castShadows.Load(jc);
            shadowStrength.Load(jc);
            offsetX.Load(jc);
            rotation.Load(jc);
        }
    }

    static class ColorExt
    {
        public static HSVColor ToHSV(this Color color)
        {
            float h, s, v;
            Color.RGBToHSV(color, out h, out s, out v);
            return new HSVColor
            {
                H = h,
                S = s,
                V = v
            };
        }
    }

    static class ObjEct
    {
        public static void Print(this object obj)
        {
            SuperController.LogMessage(obj.ToString());
        }

        public static void NullCheck(this object obj)
        {
            (obj == null).Print();
        }
    }
}