using System.Collections;
using SimpleJSON;

namespace CheesyFX
{
    public partial class LightMeUp
    {
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            jc["LightMeUp"] = GetJSON();
            needsStore = true;
            return jc;
        }
        
        JSONClass GetJSON(){
            JSONClass jc = new JSONClass();
            autoFocus.Store(jc);
            foreach (var headLight in headLights)
            {
                jc[headLight.name] = headLight.GetJSON();
            }
            return jc;
        }

        void LoadJSON(JSONClass jc)
        {
            autoFocus.Load(jc);
            foreach (var headLight in headLights)
            {
                headLight.LoadJSON(jc[headLight.name].AsObject);
            }
        }
        
        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (restorePhysical && restoreAppearance) // StoreType.Full
            {
                if (jc.HasKey("LightMeUp")){
                    JSONClass jcLightMeUp = jc["LightMeUp"].AsObject;
                    LoadJSON(jcLightMeUp);
                }
            }
        }
    }

    public static class JSONStorableParamExt
    {
        public static void Store(this JSONStorableParam jsonParam, JSONClass jc)
        {
            if (jsonParam is JSONStorableBool)
            {
                JSONStorableBool jParam = (JSONStorableBool)jsonParam;
                if (jParam.val != jParam.defaultVal) jc[jParam.name].AsBool = jParam.val;
            }
            else if (jsonParam is JSONStorableFloat)
            {
                JSONStorableFloat jParam = (JSONStorableFloat)jsonParam;
                if (jParam.val != jParam.defaultVal) jc[jParam.name].AsFloat = jParam.val;
            }
            else if (jsonParam is JSONStorableString)
            {
                JSONStorableString jParam = (JSONStorableString)jsonParam;
                if (jParam.val != jParam.defaultVal) jc[jParam.name] = jParam.val;
            }
            else if (jsonParam is JSONStorableStringChooser)
            {
                JSONStorableStringChooser jParam = (JSONStorableStringChooser)jsonParam;
                if (jParam.val != jParam.defaultVal) jc[jParam.name] = jParam.val;
            }
            else if (jsonParam is JSONStorableColor)
            {
                JSONStorableColor jParam = (JSONStorableColor)jsonParam;
                JSONArray ja = new JSONArray();
                ja[0].AsFloat = jParam.val.H;
                ja[1].AsFloat = jParam.val.S;
                ja[2].AsFloat = jParam.val.V;
                jc[jParam.name] = ja;
            }
        }
        
        public static void Load(this JSONStorableParam jsonParam, JSONClass jc)
        {
            if (jsonParam is JSONStorableBool)
            {
                JSONStorableBool jParam = (JSONStorableBool)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name].AsBool;
            }
            else if (jsonParam is JSONStorableFloat)
            {
                JSONStorableFloat jParam = (JSONStorableFloat)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name].AsFloat;
            }
            else if (jsonParam is JSONStorableString)
            {
                JSONStorableString jParam = (JSONStorableString)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name];
            }
            else if (jsonParam is JSONStorableStringChooser)
            {
                JSONStorableStringChooser jParam = (JSONStorableStringChooser)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name];
            }
            else if (jsonParam is JSONStorableColor)
            {
                JSONStorableColor jParam = (JSONStorableColor)jsonParam;
                if (jc.HasKey(jParam.name))
                {
                    HSVColor loadedColor = new HSVColor
                    {
                        H = jc[jParam.name][0].AsFloat,
                        S = jc[jParam.name][1].AsFloat,
                        V = jc[jParam.name][2].AsFloat
                    };
                    jParam.val = loadedColor;
                }
            }
        }
    }
}